/***************************************************************************
                          fft.h  -  Main FFT header file
                             -------------------
    begin                : Tue 20 November 2001
    copyright            : (C) 2001 by Gav Wood
    email                : gav@indigoarchive.net
 ***************************************************************************/

#ifndef __FFT
#define __FFT

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#pragma interface

#include <vector>
using namespace std;

#ifdef __SUCS
#include "spectrum.h"
#include "buffer.h"
#else
#include <sucs/spectrum.h>
#include <sucs/buffer.h>
#endif
using namespace sucs;

#ifndef PI
 #ifdef M_PI
  #define PI M_PI
 #else
  #define PI 3.14159265358979323846
 #endif
#endif

namespace sucs
{

class FFTSpectrum: public LinearSpectrum
{
public:
	void killTreble(float cutOff = 4000);
	
	FFTSpectrum(float *d, int size, float base = 0, bool logPowers = false) : LinearSpectrum(d, size, base, logPowers) {}
	FFTSpectrum(float base = 0, bool logPowers = false) : LinearSpectrum(base, logPowers) {}
};

template<class sampleType> class FFT
{
	unsigned bufferSizeLog, bufferSize;
	FFTSpectrum theOutput;

	vector<unsigned int> bitReverse;
	vector<float> sintable, costable, real, imag;

private:
	float theBiggest;

	int reverseBits(unsigned int initial);
	void setup(unsigned newSizeLog, float freq);

public:
	const float biggest() { return theBiggest; }

	const FFTSpectrum &perform(Buffer theInput, int max = 32768);
//	const FFTSpectrum &performNew(Buffer theInput, int max = 32768);
	const FFTSpectrum &output() { return theOutput; }
	const FFTSpectrum &operator()(const Buffer &theInput) { return perform(theInput); }

	FFT();
	FFT(unsigned newSizeLog, float freq);
	FFT(Buffer &theBuffer);
	~FFT();
};

template<class sampleType> FFT <sampleType>::FFT() : theOutput(0, true)
{
	// null setup
	bufferSize = 0;
	bufferSizeLog = 0;
}

template<class sampleType> FFT <sampleType>::~FFT()
{
}

template<class sampleType> FFT <sampleType>::FFT(unsigned newSizeLog, float freq) : theOutput(0, true)
{
	setup(newSizeLog, freq);
}

template<class sampleType> FFT <sampleType>::FFT(Buffer &theBuffer) : theOutput(0, true)
{
	setup(int(ceil(log(float(theBuffer.size())) / log(2.0) - 0.000001)), theBuffer.frequency());
}

template<class sampleType> int FFT <sampleType>::reverseBits(unsigned int initial)
{
	unsigned int reversed = 0;
	for(unsigned loop = 0; loop < bufferSizeLog; loop++)
	{
		reversed <<= 1;
		reversed += (initial & 1);
		initial >>= 1;
	}
	return reversed;
}

template<class sampleType> void FFT <sampleType>::setup(unsigned newSizeLog, float freq)
{
	bufferSizeLog = newSizeLog;
	bufferSize = 1 << newSizeLog;

	bitReverse.resize(bufferSize);
	for(unsigned i = 0; i < bufferSize; i++)
		bitReverse[i] = reverseBits(i);

	// set up sin/cos lookups
	sintable.resize(bufferSize / 2);
	costable.resize(bufferSize / 2);
	for(unsigned i = 0; i < bufferSize / 2; i++)
	{
		float j = 2 * PI * i / bufferSize;
		costable[i] = cos(j);
		sintable[i] = sin(j);
	}

	real.resize(bufferSize);
	imag.resize(bufferSize);
	
	theOutput.resize(bufferSize / 2);
	theOutput.setBaseFrequency(freq / bufferSize);
}


template<class sampleType> const FFTSpectrum &FFT <sampleType>::perform(Buffer theInput, int max)
{
	if(!bufferSize)
		setup(int(ceil(log(float(theInput.size())) / log(2.0))), theInput.frequency());

	for(unsigned i = 0; i < bufferSize; i++)
	{
		real[i] = theInput[bitReverse[i] % theInput.size()];
		real[i] /= max;
		imag[i] = 0;
	}
	
	unsigned exchanges = 1, factfact = bufferSize / 2;
	float fact_real, fact_imag, tmp_real, tmp_imag;

	for(unsigned i = bufferSizeLog; i != 0; i--)
	{
		for(unsigned j = 0; j != exchanges; j++)
		{
			fact_real = costable[j * factfact];
			fact_imag = sintable[j * factfact];

			for(unsigned k = j; k < bufferSize; k += exchanges << 1)
			{
				int k1 = k + exchanges;
				tmp_real = fact_real * real[k1] - fact_imag * imag[k1];
				tmp_imag = fact_real * imag[k1] + fact_imag * real[k1];
				real[k1] = real[k] - tmp_real;
				imag[k1] = imag[k] - tmp_imag;
				real[k]  += tmp_real;
				imag[k]  += tmp_imag;
			}
		}
		exchanges <<= 1;
		factfact >>= 1;
	}

	float theMagnitude = (bufferSize / 2);
	theMagnitude *= theMagnitude;
	for(unsigned i = 0; i < bufferSize / 2; i++)
		theOutput[i] = ((real[i] * real[i]) + (imag[i] * imag[i])) / theMagnitude;
	theOutput[0] /= 4;

	return theOutput;
}


void FFTSpectrum::killTreble(float cutOff)
{
	if(!theBaseFreq) return;
	
	int ntSize = int(cutOff / theBaseFreq);
	if(ntSize == tData.size()) return;
	
	tData.resize(ntSize);
}
/*
//-----------------------
// ALTERNATIVE FFT AREA
//-----------------------


int **gFFTBitTable = NULL;
const int MaxFastBits = 16;

int IsPowerOfTwo(int x)
{
   if (x < 2)
      return false;

   if (x & (x - 1))             // Thanks to 'byang' for this cute trick!
      return false;

   return true;
}

int NumberOfBitsNeeded(int PowerOfTwo)
{
   int i;

   if (PowerOfTwo < 2) {
      fprintf(stderr, "Error: FFT called with size %d\n", PowerOfTwo);
      exit(1);
   }

   for (i = 0;; i++)
      if (PowerOfTwo & (1 << i))
         return i;
}

int ReverseBits(int index, int NumBits)
{
   int i, rev;

   for (i = rev = 0; i < NumBits; i++) {
      rev = (rev << 1) | (index & 1);
      index >>= 1;
   }

   return rev;
}

void InitFFT()
{
   gFFTBitTable = new int *[MaxFastBits];

   int len = 2;
   for (int b = 1; b <= MaxFastBits; b++) {

      gFFTBitTable[b - 1] = new int[len];

      for (int i = 0; i < len; i++)
         gFFTBitTable[b - 1][i] = ReverseBits(i, b);

      len <<= 1;
   }
}

inline int FastReverseBits(int i, int NumBits)
{
   if (NumBits <= MaxFastBits)
      return gFFTBitTable[NumBits - 1][i];
   else
      return ReverseBits(i, NumBits);
}

void doFFT(int NumSamples,
         bool InverseTransform,
         float *RealIn, float *ImagIn, float *RealOut, float *ImagOut)
{
   int NumBits;                 // Number of bits needed to store indices
   int i, j, k, n;
   int BlockSize, BlockEnd;

   double angle_numerator = 2.0 * M_PI;
   float tr, ti;                // temp real, temp imaginary 

   if (!IsPowerOfTwo(NumSamples)) {
   	cerr << NumSamples << "is not a power of 2" << endl;
      exit(1);
   }

   if (!gFFTBitTable)
      InitFFT();

   if (InverseTransform)
      angle_numerator = -angle_numerator;

   NumBits = NumberOfBitsNeeded(NumSamples);

    //   Do simultaneous data copy and bit-reversal ordering into outputs...

   for (i = 0; i < NumSamples; i++) {
      j = FastReverseBits(i, NumBits);
      RealOut[j] = RealIn[i];
      ImagOut[j] = (ImagIn == NULL) ? 0.0 : ImagIn[i];
   }

    //   Do the FFT itself...

   BlockEnd = 1;
   for (BlockSize = 2; BlockSize <= NumSamples; BlockSize <<= 1) {

      double delta_angle = angle_numerator / (double) BlockSize;

      float sm2 = sin(-2 * delta_angle);
      float sm1 = sin(-delta_angle);
      float cm2 = cos(-2 * delta_angle);
      float cm1 = cos(-delta_angle);
      float w = 2 * cm1;
      float ar0, ar1, ar2, ai0, ai1, ai2;

      for (i = 0; i < NumSamples; i += BlockSize) {
         ar2 = cm2;
         ar1 = cm1;

         ai2 = sm2;
         ai1 = sm1;

         for (j = i, n = 0; n < BlockEnd; j++, n++) {
            ar0 = w * ar1 - ar2;
            ar2 = ar1;
            ar1 = ar0;

            ai0 = w * ai1 - ai2;
            ai2 = ai1;
            ai1 = ai0;

            k = j + BlockEnd;
            tr = ar0 * RealOut[k] - ai0 * ImagOut[k];
            ti = ar0 * ImagOut[k] + ai0 * RealOut[k];

            RealOut[k] = RealOut[j] - tr;
            ImagOut[k] = ImagOut[j] - ti;

            RealOut[j] += tr;
            ImagOut[j] += ti;
         }
      }

      BlockEnd = BlockSize;
   }

      //   Need to normalize if inverse transform...

   if (InverseTransform) {
      float denom = (float) NumSamples;

      for (i = 0; i < NumSamples; i++) {
         RealOut[i] /= denom;
         ImagOut[i] /= denom;
      }
   }
}

template<class sampleType> const FFTSpectrum &FFT <sampleType>::performNew(Buffer theInput, int max)
{
	if(!bufferSize)
		setup(int(ceil(log(float(theInput.size())) / log(2.0) - 0.000001)), theInput.frequency());
	int NumSamples = bufferSize;
	cerr << "Input size: " << theInput.size() << ", bufferSize: " << bufferSize << endl;
	

   int Half = NumSamples / 2;
   int i;

   float theta = M_PI / Half;

   float *tmpReal = new float[Half];
   float *tmpImag = new float[Half];
   float *RealOut = new float[Half];
   float *ImagOut = new float[Half];

   for (i = 0; i < Half; i++) {
      tmpReal[i] = theInput[(2 * i) % theInput.size()]; tmpReal[i] /= max;
      tmpImag[i] = theInput[(2 * i + 1) % theInput.size()]; tmpImag[i] /= max;
   }

   
   doFFT(Half, 0, tmpReal, tmpImag, RealOut, ImagOut);

   
   float wtemp = float (sin(0.5 * theta));

   float wpr = -2.0 * wtemp * wtemp;
   float wpi = float (sin(theta));
   float wr = 1.0 + wpr;
   float wi = wpi;

   int i3;

   float h1r, h1i, h2r, h2i, rt, it;

   for (i = 1; i < Half / 2; i++) {

      i3 = Half - i;

      h1r = 0.5 * (RealOut[i] + RealOut[i3]);
      h1i = 0.5 * (ImagOut[i] - ImagOut[i3]);
      h2r = 0.5 * (ImagOut[i] + ImagOut[i3]);
      h2i = -0.5 * (RealOut[i] - RealOut[i3]);

      rt = h1r + wr * h2r - wi * h2i;
      it = h1i + wr * h2i + wi * h2r;

      theOutput[i] = rt * rt + it * it;

      rt = h1r - wr * h2r + wi * h2i;
      it = -h1i + wr * h2i + wi * h2r;

      theOutput[i3] = rt * rt + it * it;

      wr = (wtemp = wr) * wpr - wi * wpi + wr;
      wi = wi * wpr + wtemp * wpi + wi;
   }
   
   rt = (h1r = RealOut[0]) + ImagOut[0];
   it = h1r - ImagOut[0];
   theOutput[0] = rt * rt + it * it;
   
   rt = RealOut[Half / 2];
   it = ImagOut[Half / 2];
   theOutput[Half / 2] = rt * rt + it * it;
   
   delete[]tmpReal;
   delete[]tmpImag;
   delete[]RealOut;
   delete[]ImagOut;
   
   return theOutput;
}

*/
}

#endif
