/***************************************************************************
                          spectrum.cpp  -  The main Spectrum implementation file
                             -------------------
    begin                : Fri 17 January 2003
    copyright            : (C) 2001 by Gav Wood
    email                : gav@indigoarchive.net
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define __SUCS 1

#pragma implementation

#include <iostream>
#include <cmath>
using namespace std;

#include "spectrum.h"
using namespace sucs;

Spectrum::Spectrum(const Spectrum &source)
{
	tData = source.tData;
	haveLogPowers = source.haveLogPowers;
	theArray = 0;
	theMaxSPL = source.theMaxSPL;
}

Spectrum &Spectrum::operator=(const Spectrum &source)
{
	killArray();
	tData = source.tData;
	haveLogPowers = source.haveLogPowers;
	theMaxSPL = source.theMaxSPL;
	return *this;
}

Spectrum::Spectrum(float *d, int size, bool logPowers)
{
	haveLogPowers = logPowers;
	tData.resize(size);
	for(int i = 0; i < tData.size(); i++)
		tData[i] = d[i];
	theArray = 0;
	theMaxSPL = 0.355655882;	// 85dB, re 20Pa
}

Spectrum::Spectrum(bool logPowers)
{
	haveLogPowers = logPowers;
	tData.resize(0);
	theArray = 0;
	theMaxSPL = 0.355655882;	// 85dB, re 20Pa
}

Spectrum::~Spectrum()
{
	if(theArray) delete [] theArray;
}

float Spectrum::diffByCosine(const Spectrum &comp) const
{
	double ret = 0, mx = 0, my = 0, s = min(size(), comp.size());
	const float *x = array(), *y = comp.array();

	for(int i = 0; i < s; i++)
	{
		ret += x[i] * y[i];
		mx += x[i] * x[i];
		my += y[i] * y[i];
	}
	float div = sqrt(mx) * sqrt(my);
	if(!isnan(div)) if(!isnan(ret / div))
		return ret / div;
	return 0;
}

void Spectrum::makeArray() const
{
	if(theArray) delete [] theArray;
	theArray = new float[tData.size()];
	for(int i = 0; i < tData.size(); i++)
		theArray[i] = tData[i];
}

void Spectrum::nearestRescale(int nSize)
{
	killArray();
	vector<float> nData;
	nData.resize(nSize);

	for(int i = 0; i < nData.size(); i++)
		nData[i] = tData[i * tData.size() / nSize];
	tData = nData;
}

void Spectrum::blockRescale(int nSize)
{
	killArray();
	float sizeCE = float(tData.size()) / float(nSize);
	vector<float> oData = tData;
	tData.resize(nSize);

	for(unsigned i = 0; i < nSize; i++)
	{
		float from = float(i) * sizeCE, to = float(i + 1) * sizeCE;
		if(ceil(from) > floor(to))		// i.e. they're the same digit
			tData[i] = sizeCE * oData[int(floor(to))];
		else
		{
			tData[i] = (int(ceil(from)) - from) * oData[int(floor(from))];
			for(int j = int(ceil(from)); j < int(floor(to)); j++)
				tData[i] += oData[j];
			tData[i] += (to - floor(to)) * oData[int(floor(to))];
		}
		tData[i] /= sizeCE;
	}
}

void Spectrum::basicRescale(int nSize)
{
	killArray();
	vector<float> oData = tData;
	tData.resize(nSize);

	if(nSize < oData.size())
	{
		for(int i = 0; i < nSize; i++) tData[i] = 0;
		for(int i = 0; i < oData.size(); i++)
			tData[i * nSize / oData.size()] += (oData[i] * nSize / oData.size());
	}
	else if(nSize > oData.size())
		for(int i = 0; i < nSize; i++)
			tData[i] = oData[i * oData.size() / nSize ];
}

unsigned Spectrum::getTopBand(bool allowZero) const
{
	unsigned f = 1;
	for(unsigned i = allowZero ? 0 : 1; i < tData.size(); i++)
		if(tData[i] > tData[f])
			f = i;
	return f;
}

void Spectrum::resize(int nSize)
{
	killArray();
	tData.resize(nSize);
}

void Spectrum::populate(float *d, int size)
{
	tData.resize(size);
	for(int i = 0; i < tData.size(); i++)
		tData[i] = d[i];
}
/*
void Spectrum::logNormalise()
{
	float logMagnitude = log(theMagnitude);
	for(unsigned i = 0; i < tData.size(); i++)
	{	tData[i] = log(tData[i] + 1);
		tData[i] /= logMagnitude;
	}
}
*/
void Spectrum::convertToDecibels(float relativeSPL, bool clipLow)
{
	killArray();
	for(unsigned i = 0; i < tData.size(); i++)
	{	if(clipLow && tData[i] * theMaxSPL < relativeSPL)
			tData[i] = 0;
		else
			tData[i] = 10 * log10(tData[i] * theMaxSPL / relativeSPL);
	}
}

void Spectrum::convertToLinear(float relativeSPL)
{
	killArray();
	for(unsigned i = 0; i < tData.size(); i++)
		tData[i] = pow(tData[i] / 10, 10) / theMaxSPL * relativeSPL;
}

void Spectrum::normaliseEnergy(bool catenateMagnitude, bool catenateOffset)
{
	killArray();
	float min = tData[0], max = tData[0];
	for(int i = 0; i < tData.size(); i++)
		if(min > tData[i]) min = tData[i];
		else if(max < tData[i]) max = tData[i];
	float mag = max - min, off = min;
	for(int i = 0; i < tData.size(); i++)
		tData[i] = (tData[i] - off) / mag;
	if(catenateMagnitude) tData.push_back(mag);
	if(catenateOffset) tData.push_back(off);
}

int Spectrum::rollOff(float amount) const
{
	float total = 0, cutoff = energy() * amount;
	for(int i = 0; i < tData.size(); i++)
	{	total += tData[i];
		if(total > cutoff)
			return i;
	}
	return tData.size();
}

float Spectrum::energy() const
{
	float total = 0;
	for(int i = 0; i < tData.size(); i++)
		total += tData[i];
	return total;
}

float Spectrum::centroid() const
{
	float num = 0, denom = 0;
	for(int i = 0; i < tData.size(); i++)
	{	num += tData[i] * i;
		denom += tData[i];
	}
	return num / denom;
}

float BandedSpectrum::centroidFrequency() const
{
	float num = 0, denom = 0;
	for(int i = 0; i < tData.size(); i++)
	{	num += tData[i] * bandFrequency(i);
		denom += tData[i];
	}
	return num / denom;
}

float BandedSpectrum::interpolate(int fBand, float fOff, int tBand, float tOff) const
{
	if(tBand != fBand)
	{	float ret = 0;
		for(int i = fBand + 1; i < tBand; i++)
			ret += tData[i];
		return ret + interpolate(fBand, fOff, fBand, 0.5) + interpolate(tBand, -0.5, tBand, tOff);
	}
	else
		return tData[fBand] * (tOff - fOff);
}

float BandedSpectrum::getPowerInFrequencyBand(float from, float to) const
{
	int fBand = -1, tBand = -1;
	float fOff, tOff;
	for(int band = 0; band < tData.size() && (fBand == -1 || tBand == -1); band++)
	{	float bFrom, bTo;
		if(!band) bFrom = 0; else bFrom = (bandFrequency(fBand) + bandFrequency(fBand - 1)) / 2;
		if(band == tData.size() - 1) bTo = bandFrequency(fBand) * 2 - bandFrequency(fBand - 1); else bTo = (bandFrequency(fBand) + bandFrequency(fBand + 1)) / 2;		if(from >= bFrom && from < bTo) { fBand = band; fOff = ((from - bFrom) / (bTo - bFrom)) - 0.5; }
		if(to >= bFrom && to < bTo) { tBand = band; tOff = ((to - bFrom) / (bTo - bFrom)) - 0.5; }
	}
	return interpolate(fBand, fOff, tBand, tOff);
}

void BandedSpectrum::convertToPhons()
{
	killArray();
	const int numEqBands = 22;
	const int numCurves = 6;
	static float freqs[numEqBands] = { 31.62, 50, 70.7, 100, 141.4, 200, 316.2, 500, 707.1, 1000, 1414, 1682, 2000, 2515, 3162, 3976, 5000, 7071, 10000, 11890, 14140, 15500 };
	static float curves[numCurves][numEqBands] = {
		{  55,   40,   32,   24,   19,   14,   10,    6,    4,    3,    2,    2,    0,   -2,   -5,   -4,    0,    5,   10,   14,   25,   35 },
		{  66,   52,   43,   37,   32,   27,   23,   21,   20,   20,   20,   20,   19,   16,   13,   13,   18,   22,   25,   30,   40,   50 },
		{  76,   64,   57,   51,   47,   43,   41,   41,   40,   40,   40,   39.5, 38,   35,   33,   33,   35,   41,   46,   50,   60,   70 },
		{  89,   79,   74,   70,   66,   63,   61,   60,   60,   60,   60,   59,   56,   53,   52,   53,   56,   61,   65,   70,   80,   90 },
		{ 103,   96,   92,   88,   85,   83,   81,   80,   80,   80,   80,   79,   76,   72,   70,   70,   75,   79,   83,   87,   95,  105 },
		{ 118,  110,  107,  105,  103,  102,  101,  100,  100,  100,  100,   99,   97,   94,   90,   90,   95,  100,  103,  105,  108,  115 } };

	for(int band = 0; band < tData.size(); band++)
	{
//		cerr << "Band " << band << endl;

		// find two bands it's between.
		int bandA, bandB;
		float fracA, fracB;
		float f = bandFrequency(band);
		if(f < freqs[0])
		{	bandA = bandB = 0;
			fracA = fracB = 0.5;
		}
		else if(f >= freqs[numEqBands - 1])
		{	bandA = bandB = numEqBands - 1;
			fracA = fracB = 0.5;
		}
		else
			for(bandA = 0, bandB = 1; bandB < numEqBands; bandA++, bandB++)
				if(f >= freqs[bandA] && f < freqs[bandB])
				{	fracB = (f - freqs[bandA]) / (freqs[bandB] - freqs[bandA]);
					fracA = 1 - fracB;
					break;
				}
//		cerr << "Between: " << bandA << " and " << bandB << " (" << fracA << ", " << fracB << ")" << endl;
		// now, go through all curves and interpolate accordingly to find the two that enclose the point
		int curveA, curveB;
		float fracCA, fracCB;
		float p = tData[band];

//		cerr << "Band power: " << p << endl;
//		cerr << "Lower limit: " << (curves[0][bandA] * fracA + curves[0][bandB] * fracB) << endl;
//		cerr << "Upper limit: " << (curves[numCurves - 1][bandA] * fracA + curves[numCurves - 1][bandB] * fracB) << endl;
		if(p < curves[0][bandA] * fracA + curves[0][bandB] * fracB)
		{	curveA = curveB = 0;
			fracCA = fracCB = 0.5;
		}
		else if(p >= curves[numCurves - 1][bandA] * fracA + curves[numCurves - 1][bandB] * fracB)
		{	curveA = curveB = numCurves - 1;
			fracCA = fracCB = 0.5;
		}
		else
		{
//			cerr << "Attempting to find..." << endl;
			for(curveA = 0, curveB = 1; curveB < numCurves; curveA++, curveB++)
			{
//				cerr << "Limits (" << p << "): " << (curves[curveA][bandA] * fracA + curves[curveA][bandB] * fracB);
//				cerr << " --- " << (curves[curveB][bandA] * fracA + curves[curveB][bandB] * fracB) << endl;

				if(p >= curves[curveA][bandA] * fracA + curves[curveA][bandB] * fracB
					&& p < curves[curveB][bandA] * fracA + curves[curveB][bandB] * fracB)
				{
//					cerr << "Stopped at " << curveA << ", " << curveB << endl;
					fracCB = (p - (curves[curveA][bandA] * fracA + curves[curveA][bandB] * fracB))
						/ ((curves[curveB][bandA] * fracA + curves[curveB][bandB] * fracB) - (curves[curveA][bandA] * fracA + curves[curveA][bandB] * fracB));
					fracCA = 1 - fracCB;
					break;
				}
			}
		}
//		cerr << "Between curves: " << curveA << " and " << curveB << " (" << fracCA << ", " << fracCB << ")" << endl;
		float phons = 20 * (curveA * fracCA + curveB * fracCB);
		tData[band] = phons;
	}
}

void BandedSpectrum::convertToSones()
{
	killArray();
	for(int band = 0; band < tData.size(); band++)
		tData[band] = pow(2, (tData[band] - 40) / 10);
}

FreeSpectrum::FreeSpectrum(const BandedSpectrum &source) : BandedSpectrum(source)
{
	theBandFreqs.resize(tData.size());
	for(int i = 0; i < tData.size(); i++)
		theBandFreqs[i] = source.bandFrequency(i);
}

FreeSpectrum &FreeSpectrum::operator=(BandedSpectrum &source)
{
	Spectrum::operator=(source);
	theBandFreqs.resize(source.size());
	for(int i = 0; i < source.size(); i++)
		theBandFreqs[i] = source.bandFrequency(i);
}

void FreeSpectrum::correctForScale(unsigned nSize)
{
	killArray();
	float z = theBandFreqs.size() / nSize;
	vector<float> oFreqs = theBandFreqs;
	for(int i = 0; i < nSize; i++)
		theBandFreqs[i] = oFreqs[int(round(float(i) * z))];
}

void FreeSpectrum::templateRescale(const Template &theTemplate)
{
	killArray();
	vector<float> nData, nFreqs;
	nData.resize(theTemplate.bands());
	nFreqs.resize(theTemplate.bands());

	for(int i = 0; i < nFreqs.size(); i++)
	{	nFreqs[i] = theTemplate.band(i);
		float mag = 0;
		for(int j = 0; j < theTemplate.components(i); j++)
		{	nData[i] += tData[theTemplate.componentBand(i, j)] * theTemplate.componentFraction(i, j);
			mag += theTemplate.componentFraction(i, j);
		}
//		cerr << "Should divide by " << mag << endl;
//		if(mag) nData[i] /= mag;
	}

	tData = nData;
	theBandFreqs = nFreqs;
}

void FreeSpectrum::geometricRescale(unsigned nSize, bool inverse)
{
	killArray();
	if(nSize > tData.size()) return;
	vector<float> nData, nFreqs;
	nData.resize(nSize);
	nFreqs.resize(nSize);
	float m = tData.size(), n = nSize;
	float t = (n / 2) / (m / n - 1);

	float tot = 0;
	for(unsigned i = 0; i < n; i++)
	{	float l = inverse ? (n - float(i) - 1.0) : float(i);
		l = l / t + 1;
		nData[i] = nFreqs[i] = 0;
		for(int j = int(floor(tot)); j < int(floor(tot + l)); j++)
		{	nData[i] += tData[j];
			nFreqs[i] += theBandFreqs[j];
		}
		nData[i] /= int(floor(tot + l)) - int(floor(tot));
		nFreqs[i] /=  int(floor(tot + l)) - int(floor(tot));
		tot += l;

	}
	tData = nData;
	theBandFreqs = nFreqs;
}

vector<FreeSpectrum> FreeSpectrum::templateSplit(const Template &theTemplate)
{
	killArray();
	vector<FreeSpectrum> newSpectra(theTemplate.bands());
	for(int i = 0; i < newSpectra.size(); i++)
		for(int j = 0; j < theTemplate.components(i); j++)
		{	newSpectra[i].tData.push_back(tData[theTemplate.componentBand(i, j)] * theTemplate.componentFraction(i, j));
			newSpectra[i].theBandFreqs.push_back(theBandFreqs[theTemplate.componentBand(i, j)]);
		}
	return newSpectra;
}

// broken
/*
void FreeSpectrum::exponentialRescale(unsigned nSize, bool inverse)
{
	if(nSize > tData.size()) return;
	vector<float> nData, nFreqs;
	nData.resize(nSize);
	nFreqs.resize(nSize);
	float m = tData.size(), n = nSize;
	float r = pow(m + 1, 1 / n);
	cerr << "r = " << r << ", m = " << m << ", n = " << n << endl;
	for(unsigned j = 0; j < n; j++)
	{
		unsigned i = j;
		float from = pow(r, float(j)) - 1, to = pow(r, float(j + 1)) - 1;
		if(inverse)
		{	i = int(n) - j - 1;
			float f = from; from = to; to = f;
			from = float(m) - from - 1;
			to = float(m) - to - 1;
		}
		cerr << "i = " << i << ", from = " << from << ", to = " << to << endl;
		nData[i] = (int(ceil(from)) - from) * tData[int(floor(from))];
		for(int j = int(ceil(from)); j < int(floor(to)); j++)
			nData[i] += tData[j];
		nData[i] += (to - floor(to)) * tData[int(floor(to))];
		nData[i] /= (to - from);
		nFreqs[i] = theBandFreqs[int(round((from + to) / 2))];
	}
	tData = nData;
	theBandFreqs = nFreqs;
}
*/

void Template::setBands(int bands)
{
	theData.clear();
	theData.resize(bands);
	theBands.resize(bands);
}

int Template::findBandOfFrequency(float freq) const
{
	for(int b = 0; b < theBandEdges.size(); b++)
		if(freq >= theBandEdges[b].first && freq < theBandEdges[b].second)
			return b;
}

void Template::addFrequencyBand(int band, float from, float to, const BandedSpectrum &theSpectrum)
{
	int fBand = -1, tBand = -1;
	float fOff, tOff;

	theBandEdges.push_back(pair<float,float>(from, to));

	for(int b = 0; b < theSpectrum.size() && (fBand == -1 || tBand == -1); b++)
	{	float bFrom, bTo;
		if(!b) bFrom = 0; else bFrom = (theSpectrum.bandFrequency(b) + theSpectrum.bandFrequency(b - 1)) / 2;
		if(b == theSpectrum.size() - 1) bTo = theSpectrum.bandFrequency(b) * 2 - theSpectrum.bandFrequency(b - 1); else bTo = (theSpectrum.bandFrequency(b) + theSpectrum.bandFrequency(b + 1)) / 2;
		if(from >= bFrom && from < bTo) { fBand = b; fOff = ((from - bFrom) / (bTo - bFrom)) - 0.5; }
		if(to >= bFrom && to < bTo) { tBand = b; tOff = ((to - bFrom) / (bTo - bFrom)) - 0.5; }
	}

	if(fBand == -1 && tBand == -1)
		return;
	if(fBand == -1 && tBand > -1)
	{	fBand = 0;
		fOff = -0.5;
	}
	if(tBand == -1 && fBand > -1)
	{	tBand = theSpectrum.size() - 1;
		tOff = 0.5;
	}

	float thisLength = 0;
	if(tBand != fBand)
	{	theData[band].push_back(pair<int, float>(fBand, 0.5 - fOff));
		thisLength += 0.5 - fOff;
		for(int i = fBand + 1; i < tBand; i++)
		{	theData[band].push_back(pair<int, float>(i, 1));
			thisLength++;
		}
		theData[band].push_back(pair<int, float>(tBand, 0.5 + tOff));
		thisLength += 0.5 + tOff;
	}
	else
	{	theData[band].push_back(pair<int, float>(tBand, tOff - fOff));
		thisLength += tOff - fOff;
	}
	if(thisLength > theLargest) theLargest = thisLength;
}

void BarkTemplate::concrete(const BandedSpectrum &theSpectrum)
{
	setBands(24);
	theLargest = 0;
	addFrequencyBand(0, 0, 100, theSpectrum);
	theBands[0] = 50;	// addFrequencyBand looks after theBandEdges
	addFrequencyBand(1, 100, 200, theSpectrum);
	theBands[1] = 150;
	addFrequencyBand(2, 200, 300, theSpectrum);
	theBands[2] = 250;
	addFrequencyBand(3, 300, 400, theSpectrum);
	theBands[3] = 350;
	addFrequencyBand(4, 400, 510, theSpectrum);
	theBands[4] = 450;
	addFrequencyBand(5, 510, 630, theSpectrum);
	theBands[5] = 570;
	addFrequencyBand(6, 630, 770, theSpectrum);
	theBands[6] = 700;
	addFrequencyBand(7, 770, 920, theSpectrum);
	theBands[7] = 840;
	addFrequencyBand(8, 920, 1080, theSpectrum);
	theBands[8] = 1000;
	addFrequencyBand(9, 1080, 1270, theSpectrum);
	theBands[9] = 1170;
	addFrequencyBand(10, 1270, 1480, theSpectrum);
	theBands[10] = 1370;
	addFrequencyBand(11, 1480, 1720, theSpectrum);
	theBands[11] = 1600;
	addFrequencyBand(12, 1720, 2000, theSpectrum);
	theBands[12] = 1850;
	addFrequencyBand(13, 2000, 2320, theSpectrum);
	theBands[13] = 2150;
	addFrequencyBand(14, 2320, 2700, theSpectrum);
	theBands[14] = 2500;
	addFrequencyBand(15, 2700, 3150, theSpectrum);
	theBands[15] = 2900;
	addFrequencyBand(16, 3150, 3700, theSpectrum);
	theBands[16] = 3400;
	addFrequencyBand(17, 3700, 4400, theSpectrum);
	theBands[17] = 4000;
	addFrequencyBand(18, 4400, 5300, theSpectrum);
	theBands[18] = 4800;
	addFrequencyBand(19, 5300, 6400, theSpectrum);
	theBands[19] = 5800;
	addFrequencyBand(20, 6400, 7700, theSpectrum);
	theBands[20] = 7000;
	addFrequencyBand(21, 7700, 9500, theSpectrum);
	theBands[21] = 8500;
	addFrequencyBand(22, 9500, 12000, theSpectrum);
	theBands[22] = 10500;
	addFrequencyBand(23, 12000, 15500, theSpectrum);
	theBands[23] = 13500;
}

