/***************************************************************************
                          spectrum.h  -  Main Spectrum header file
                             -------------------
    begin                : Tue 20 November 2001
    copyright            : (C) 2001 by Gav Wood
    email                : gav@indigoarchive.net
 ***************************************************************************/

#ifndef __SPECTRUM
#define __SPECTRUM

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#pragma interface

#include <vector>
//#include <pair>
using namespace std;

namespace sucs
{

class Spectrum
{
protected:
	vector<float> tData;
	mutable float *theArray;
	bool haveLogPowers;
	float theMaxSPL;

public:
	void setMaxSPL(float newMaxSPL) { theMaxSPL = newMaxSPL; }
	float maxSPL() { return theMaxSPL; }
	void convertToDecibels(float relativeSPL = 0.00002, bool clipLow = true);
	void convertToLinear(float relativeSPL = 0.00002);

	int rollOff(float amount = 0.85) const;
	float energy() const;
	float centroid() const;
	void normaliseEnergy(bool catenateMagnitude = false, bool catenateOffset = false);

	const float &operator[](int i) const { return tData[i % tData.size()]; }
	float &operator[](int i) { return tData[i % tData.size()]; }
	void populate(float *d, int size);

	const float *array(bool forceRemake = 0) const { if(!theArray || forceRemake) makeArray(); return theArray; }
	void makeArray() const;
	void killArray() const { if(theArray) delete [] theArray; theArray = 0; }

	float diffByCosine(const Spectrum &comp) const;

	void resize(int nSize);
	int size() const { return tData.size(); }

	void nearestRescale(int nSize);		// worst
	void basicRescale(int nSize);		// marginally better
	void blockRescale(int nSize);		// much better
//	void linearRescale(int nSize);		// best, but not implemented yet

	// TODO: obsolete
//	void wantLogPowers();
//	void wantBasePowers();
	void setLogs(bool logPowers = true) { haveLogPowers = logPowers; }

	unsigned getTopBand(bool allowZero = false) const;

	Spectrum(const Spectrum &source);
	Spectrum &operator=(const Spectrum &source);

	Spectrum(float *d, int size, bool logPowers = false);
	Spectrum(bool logPowers = false);
	virtual ~Spectrum();
};

class BandedSpectrum: public Spectrum
{
protected:
	float interpolate(int fBand, float fOff, int tBand, float tOff) const;

public:
	virtual float bandFrequency(unsigned band) const = 0;
	virtual float bandTime(unsigned band) const = 0;
	virtual void correctForScale(unsigned nSize) = 0;
	float bandBPM(unsigned band) { return bandFrequency(band) * 60; }

	// Rescale functions
	void nearestRescale(int nSize) { correctForScale(nSize); Spectrum::nearestRescale(nSize); }
	void basicRescale(int nSize) { correctForScale(nSize); Spectrum::basicRescale(nSize); }
	void blockRescale(int nSize) { correctForScale(nSize); Spectrum::blockRescale(nSize); }
//	void linearRescale(int nSize) { correctForScale(nSize); Spectrum::linearRescale(nSize); }

	// Psychoacoustic conversions
	/**
	 * Assumes source is in dB re 2Pa
	 */
	void convertToPhons();

	/**
	 * Assumes source is in phons
	 */
	void convertToSones();

	float centroidFrequency() const;

	float getTopFrequency(bool allowZero = false) const { return bandFrequency(getTopBand(allowZero)); }
	float getTopTime(bool allowZero = false) const { return bandTime(getTopBand(allowZero)); }

	float getPowerInFrequencyBand(float from, float to) const;

	BandedSpectrum(float *d, int size, bool logPowers = false) : Spectrum(d, size, logPowers) {}
	BandedSpectrum(bool logPowers = false) : Spectrum(logPowers) {}
	BandedSpectrum(const Spectrum &source) : Spectrum(source) {}
};

class Template
{
protected:
	vector< vector< pair<int, float> > > theData;
	vector<float> theBands;
	vector< pair<float, float> > theBandEdges;
	float theLargest;

	void addFrequencyBand(int band, float from, float to, const BandedSpectrum &theSpectrum);
	void setBands(int bands);

public:
	virtual void concrete(const BandedSpectrum &theSpectrum) = 0;
	float largest() const { return theLargest; }
	int findBandOfFrequency(float freq) const;

	inline int bands() const { return theBands.size(); }
	inline float band(int band) const { return theBands[band]; }
	inline int components(int band) const { return theData[band].size(); }
	inline int componentBand(int band, int index) const { return theData[band][index].first; }
	inline float componentFraction(int band, int index) const { return theData[band][index].second; }
};

class BarkTemplate: public Template
{
public:
	virtual void concrete(const BandedSpectrum &theSpectrum);
};

class FreeSpectrum: public BandedSpectrum
{
protected:
	vector<float> theBandFreqs;

public:
	float bandFrequency(unsigned band) const { return theBandFreqs[band]; }
	float bandTime(unsigned band) const { return 1 / theBandFreqs[band]; }
	void correctForScale(unsigned nSize);

	// Conversion functions
	//void exponentialRescale(unsigned nSize, bool inverse = false);	// broken
	void geometricRescale(unsigned nSize, bool inverse = false);
	void templateRescale(const Template &theTemplate);

	// Spliting functions
	vector<FreeSpectrum> templateSplit(const Template &theTemplate);

	FreeSpectrum &operator=(BandedSpectrum &source);

	FreeSpectrum(float *d, int size, bool logPowers = false) : BandedSpectrum(d, size, logPowers) {}
	FreeSpectrum(bool logPowers = false) : BandedSpectrum(logPowers) {}
	FreeSpectrum(const BandedSpectrum &source);
};

class LinearSpectrum: public BandedSpectrum
{
protected:
	float theBaseFreq;

public:
	float bandFrequency(unsigned band) const { return theBaseFreq * band; }
	float bandTime(unsigned band) const { return 1 / (theBaseFreq * band); }

	void setBaseFrequency(float newBaseFreq) { theBaseFreq = newBaseFreq; }
	void correctForScale(unsigned nSize) { theBaseFreq = theBaseFreq / nSize * tData.size(); }
	// ***************
	// wrong???
	// theBaseFreq = theBaseFreq * nSize / tData.size(); is probably not the case!!!
	// ***************

	LinearSpectrum(float *d, int size, float base = 0, bool logPowers = false) : BandedSpectrum(d, size, logPowers) { theBaseFreq = base; }
	LinearSpectrum(float base = 0, bool logPowers = false) : BandedSpectrum(logPowers) { theBaseFreq = base; }
};

class InverseSpectrum: public BandedSpectrum
{
protected:
	float theBaseTime;

public:
	float bandTime(unsigned band) const { return theBaseTime * (tData.size() - band); }
	float bandFrequency(unsigned band) const { return 1.0 / (theBaseTime * (tData.size() - band)); }

	void setBaseTime(float newBaseTime) { theBaseTime = newBaseTime; }
	void correctForScale(unsigned nSize) { theBaseTime = theBaseTime / nSize * tData.size(); }

	InverseSpectrum(float *d, int size, float base = 0, bool logPowers = false) : BandedSpectrum(d, size, logPowers) { theBaseTime = base; }
	InverseSpectrum(float base = 0, bool logPowers = false) : BandedSpectrum(logPowers) { theBaseTime = base; }
};

}

#endif
